// HLSL shader for Nauhuri
// copyright Keijo "Kegetys" Ruotsalainen
// http://www.kegetys.fi

#define COLOR_WATER float4(0.2,0.3,0.9,1.0)
#define COLOR_HORIZON float4(0.84f, 0.9f, 1.0f, 1)
#define COLOR_SKY float4(0.2f, 0.3f, 0.5f, 1)
#define COLOR_SEABOTTOM  float4(0.2, 0.2, 0.2, 1)
#define COLOR_BEACH float4(0.87, 0.82, 0.65, 1)
#define COLOR_LAND float4(0.59, 0.56, 0.51, 1)
#define COLOR_UWATER float4(0.0,0.3,1.0,1.0)

// VS registers
float4x4 MatWorld : register(c0); 
float4x4 MatViewProj : register(c4);
float4x4 MatViewProjSB : register(c8);
float4   camPos : register(c12); 

// PS registers
float4 matColor : register(c0); 

// Shared registers
float4 renderClip : register(c128); // (width, height, znear, zfar)
float4 renderFade : register(c127); // (fadeNear, fadeFar, x, x)

struct PSO {
    float4 col   : COLOR0;
};

struct VSI {
    float4 pos   : POSITION;
	float4 norm  : NORMAL;
};

struct VSO {
    float4 pos    : POSITION;
	float4 col    : COLOR0;
	float4 chaze  : COLOR1;
	float4 posPS  : TEXCOORD0; // for PS: pre-transformed position
	float4 posPST : TEXCOORD1; // for PS: transformed position
};

void applyUnderwater(in out float4 col)
{
	if(camPos.y < 0)
	{
		col = COLOR_UWATER;
	}
}

float4 applyHaze(in float4 pos)
{	
	float dd = pos.z+renderClip.z-1500;
	dd = max(min(4000, dd), 0);	
	
	float alpha = saturate((pos.z-renderFade.x) / (renderFade.y-renderFade.x)); // Distance fade	
	return float4(dd/19000 * float3(0.7, 0.75, 0.96), 1-alpha);
}

VSO vsCommon(in VSI In)
{
	VSO Out;
	Out.posPS = In.pos;
	float4 wpos = mul(In.pos, MatWorld);
	float4 N = In.norm;
	Out.pos = mul(wpos, MatViewProj);

	// Very simple "lighting"
	Out.col = float4(N.y, N.y, N.y, 1); 
	Out.col = Out.col - (N.x*0.2);

	applyUnderwater(Out.col);	
	Out.posPST = Out.pos;
	
	Out.chaze = applyHaze(Out.pos);
	
	return Out;
};

float grid3D(float4 pos)
{
	float size = 0.1;
	float gwidth = size/2;
	if(
	  (pos.x%size < gwidth && pos.z%size < gwidth)  
	  || (pos.y%size < gwidth && pos.x%size < gwidth && pos.z%size > gwidth)  
	  || (pos.y%size < gwidth && pos.z%size < gwidth && pos.x%size > gwidth)
	 )
	 return 0.97;
	
	return 1;
}

PSO psObject(in VSO In)
{
	PSO o;
	o.col = matColor;
	o.col = o.col + (In.col*0.25);
	
	// Alpha-blend distant terrain
	float z = (In.posPST.z/renderClip.w);
	o.col.a = saturate(matColor.a-(pow(z,2)));
	
	o.col.xyz *= grid3D(In.posPS);	
	
	o.col.xzy += In.chaze;	
	o.col.a *= In.chaze.a;
	return o;
}
PSO psWater(in VSO In)
{
	PSO o;
	o.col = COLOR_WATER;
	o.col. a = 0.88f;
	o.col.xyz += In.chaze;
	return o;
}

// GROUND
VSO vsGround(in VSI In)
{
	float y = In.pos.y;
	VSO o = vsCommon(In);
	o.col = float4(0.5, 0.5, 0.5, 1);
	
	o.col -= (In.norm.x*In.norm.x/1.5f);
	o.col += In.norm.y/10.0f;
		
	applyUnderwater(o.col);	
	o.chaze = applyHaze(o.pos);		
	return o;
}

float grid(float4 pos, float z,float size, float fade, float w, float minw)
{
	const float transp = 0.05;	
	const float gwidth = max(z*(25*w), minw);
	if((pos.x+gwidth/2)%size < gwidth || (pos.z+gwidth/2)%size < gwidth)
	{
		return 1-(saturate(1-z*fade)*transp);
	}
	return 1;
}

PSO psGround(in VSO In)
{
	PSO o;
	float4 clight = In.col;
	
	const float4 cSeaBottom = COLOR_SEABOTTOM;
	const float4 cBeach = COLOR_BEACH;
	const float4 cLand = COLOR_LAND;

	// Interpolate colors
	float y = In.posPS.y;	
	o.col = lerp(cSeaBottom, cBeach, saturate((y+30)/30));
	o.col = lerp(o.col, cLand, saturate((y+1)/5));

	// Apply VS light
	o.col.xyz += clight-0.5;

	const float elev = 100;
	const float ewidth = 50;
	if(In.posPS.y%elev < ewidth)
	{
		o.col.xyz *= 0.96;
	}	

	// Alpha-blend distant terrain
	float z = (In.posPST.z/renderClip.w);
	o.col.a = saturate(1-(pow(z,8)));

	// Grid squares
	o.col.xyz *= grid(In.posPS, z, 1000, 0, 1.0, 0.25);
	o.col.xyz *= grid(In.posPS, z, 100, 1, 0.75, 0.25);
	o.col.xyz *= grid(In.posPS, z, 10, 4, 0.5, 0.25);
	o.col.xyz *= grid(In.posPS, z, 1, 32, 0.1, 0.05);
	o.col.xyz *= grid(In.posPS, z, 0.1, 128, 0.001, 0.01);
	
	o.col.xyz += In.chaze;
	//o.col.a *= In.chaze.a;
	
	return o;
}

PSO psGroundMap(in VSO In)
{
	PSO o;
	float4 clight = In.col;
	
	const float4 cSeaBottom = COLOR_SEABOTTOM;
	const float4 cBeach = COLOR_BEACH;
	const float4 cLand = COLOR_LAND;
	
	// Interpolate colors
	float y = In.posPS.y;	
	o.col = lerp(cSeaBottom, cBeach, saturate((y+30)/30));
	o.col = lerp(o.col, cLand, saturate((y+1)/5));

	// Elevation lines
	const float elev = 5;
	const float ewidth = 0.2;
	if(In.posPS.y%elev < ewidth)
	{
		o.col.xyz *= 0.8;
	}	
	
	// Apply VS light
	o.col.xyz += clight-0.5;
	o.col.a = 1;
	
	return o;
}


// SKYBOX
VSO vsSkybox(in VSI In)
{
	// Skip world transform, draw around camera. Color from Y coord
	VSO Out;
	Out.posPS = In.pos;
	Out.pos = mul(In.pos, MatViewProjSB);
	if(In.pos.y <=0)
	{
		Out.col = COLOR_WATER;
	}
	else
	{
		Out.col = lerp(COLOR_HORIZON, COLOR_SKY, min(In.pos.y*30,1.5));
	}
	
	applyUnderwater(Out.col);	
	Out.pos.z = 0.1;
	Out.posPST = Out.pos;	
	Out.chaze = float4(0,0,0,0);
	return Out;
}

PSO psSkybox(in VSO data)
{
	PSO o;
	o.col = data.col;
	return o;
}

// UI stuff etc
struct VSIcolor {
    float4 pos   : POSITION;
	float4 col   : COLOR0;
};

VSO vsColor(in VSIcolor In)
{
	VSO o = vsCommon(In);
	o.col = In.col;
	return o;
};
PSO psColor(in VSO In)
{
	PSO o;
	o.col = In.col;
	return o;
}
